/*  Test of the gtk editor widget.
 *  Copyright (C) 1998 Thomas Mailund <mailund@daimi.au.dk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include "gtkeditor.h"
#include "gtkhint.h"

#ifdef HAVE_LIBGUILE
#include <guile/gh.h>
#endif

#define BUF_MAX_SIZE 128	/* seems fair for one line */

/* global vars */
static GtkEditor *mainbuffer;
static GtkEntry *minibuffer;
static void (*minibuffer_callback)(char *) = NULL;

#ifdef HAVE_LIBGUILE
static GtkEditor *guilebuffer = NULL;
#endif

/* stuff for hints */
struct proto { char *match, *full_proto; };

struct proto *db;
int db_alloc, db_size;

int 
hint_cmp_entries (const void *p1, const void *p2)
{
  const struct proto *e2 = p2;

  return strcmp (p1, e2->match);
}

void
hint_load_db (char *filename)
{
  FILE *fp;
  int alloc, i;
  char line[1024], *entry, *tab;

  fp = fopen (filename, "r");
  if (!fp) {
    db = NULL;
    return;
  }

  alloc = 1024;
  db = malloc (alloc * sizeof (struct proto));
  i = 0;
  while (fgets (line, 1024, fp)) {
    if (i == alloc) {
      alloc *= 2;
      db = realloc (db, alloc * sizeof (struct proto));
    }
    entry = strdup (line);
    tab = strchr (entry, '\t');
    db[i].match = entry;
    db[i].full_proto = tab + 1;
    *tab = 0;
    i++;
  }
  db_alloc = alloc;
  db_size = i;
}

char *hint_callback (char *match, gpointer data)
{
  struct proto *entry;

  if (!db)
    return NULL;

  entry = bsearch (match, db, db_size,
		   sizeof (struct proto), hint_cmp_entries);

  if (entry != NULL)
    return entry->full_proto;
  return NULL;
}

/* minibuffer stuff */
static gint
minibuf_key_press (GtkWidget *widget, GdkEventKey *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->keyval == GDK_Return) {
    gtk_widget_grab_focus (GTK_WIDGET (mainbuffer));
    if (minibuffer_callback) {
      /* The callback is called with a copy of the text. It needs to
       * take care of freeing the text, should that be done. */
      minibuffer_callback (g_strdup(gtk_entry_get_text (minibuffer)));
    }
    gtk_entry_set_text (minibuffer, "");
  }

  return FALSE;
}

static void
get_text_from_minibuffer (GtkSignalFunc callback)
{
  gtk_entry_set_text (minibuffer, "");
  minibuffer_callback = callback;
  gtk_widget_grab_focus (GTK_WIDGET (minibuffer));
}

/* callbacks */

static void
search_from_point_cb (char *string)
{
#ifdef DEBUG
  g_print ("search_from_point_cb (%s)\n", string);
#endif
  gtk_editor_search_from_point (mainbuffer, string, FALSE);
  g_free (string);
}

static void
search_back_from_point_cb (char *string)
{
#ifdef DEBUG
  g_print ("search_back_from_point_cb (%s)\n", string);
#endif
  gtk_editor_search_back_from_point (mainbuffer, string, FALSE);
  g_free (string);
}

static void
regex_search_from_point_cb (char *string)
{
#ifdef DEBUG
  g_print ("regex_search_from_point_cb (%s)\n", string);
#endif
  gtk_editor_regex_search_from_point (mainbuffer, string);
  g_free (string);
}

static void
regex_search_back_from_point_cb (char *string)
{
#ifdef DEBUG
  g_print ("regex_search_back_from_point_cb (%s)\n", string);
#endif
  gtk_editor_regex_search_back_from_point (mainbuffer, string);
  g_free (string);
}


static void
load_file_cb (char *filename)
{
  int i;
  char buf[BUF_MAX_SIZE];
  FILE *fp = fopen (filename, "r");

#ifdef DEBUG
  g_print ("load_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  /* clear old text */
  i = gtk_text_get_length (GTK_TEXT(mainbuffer));
  if (i>0) {
    gtk_text_set_point (GTK_TEXT(mainbuffer), i);
    gtk_text_backward_delete (GTK_TEXT(mainbuffer), i);
  }

  /* load new */
  gtk_text_freeze (GTK_TEXT (mainbuffer));
  
  do {
    if (fgets (buf, BUF_MAX_SIZE, fp) == NULL) {
      break;
    } else {
      gtk_text_insert (GTK_TEXT (mainbuffer), NULL, NULL, NULL,
		       buf, strlen (buf));
    }
  } while (!feof (fp));

  gtk_text_thaw (GTK_TEXT (mainbuffer));

  /* clean up and hilite */
  fclose (fp);
  g_free (filename);
  /* FIXME: couldn't this be done while inserting? */
  gtk_text_set_point (GTK_TEXT (mainbuffer), 0);
  gtk_editor_hilite_screen (mainbuffer);
}


static void
save_file_cb (char *filename)
{
  int i, len;
  char *buf;
  FILE *fp = fopen (filename, "w");

#ifdef DEBUG
  g_print ("save_file_cb (%s)\n", filename);
#endif

  if (!fp) return;

  gtk_text_freeze (GTK_TEXT (mainbuffer));

  /* get_chars allocs a new buffer each time, so I have to do all this
   * g_free stuff */  
  len = gtk_text_get_length (GTK_TEXT (mainbuffer));
  for (i = 0; i < len - BUF_MAX_SIZE;) {
    buf = gtk_editable_get_chars (GTK_EDITABLE (mainbuffer), i, i + BUF_MAX_SIZE);
    fputs (buf, fp);
    i += BUF_MAX_SIZE;
    g_free (buf);
  }
  /* and the leftovers */
  buf = gtk_editable_get_chars (GTK_EDITABLE (mainbuffer), i, len);
  fputs (buf, fp);
  g_free (buf);

  gtk_text_thaw (GTK_TEXT (mainbuffer));

  fclose (fp);
}


/* FIXME: this should be done better! */
/* FIXME: these functions should be bound to keys when the keybinding
 * parser is done. */
static void
search_from_point (void)
{
  get_text_from_minibuffer (search_from_point_cb);
}

static void
search_back_from_point (void)
{
  get_text_from_minibuffer (search_back_from_point_cb);
}

static void
regex_search_from_point (void)
{
  get_text_from_minibuffer (regex_search_from_point_cb);
}

static void
regex_search_back_from_point (void)
{
  get_text_from_minibuffer (regex_search_back_from_point_cb);
}

static void
clear_buffer (void)
{
  gtk_editable_delete_text (GTK_EDITABLE (mainbuffer), 0, -1);
}

static void
load_file (void)
{
  get_text_from_minibuffer (load_file_cb);
}

static void
save_file (void)
{
  get_text_from_minibuffer (save_file_cb);
}

#ifdef HAVE_LIBGUILE
/* guilestuff */
SCM
err_handler (void *data, SCM tag, SCM throw_args)
{
  /* do clever stuff to avoid crash */
  g_print ("error\n");
  return SCM_BOOL_T;
}


static gint
guile_buffer_key_press (GtkWidget *widget, GdkEventKey *event)
{
  gchar *command;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->state & GDK_CONTROL_MASK) {
    switch (event->keyval) {
    case 'x':
      /* send buffer to intepreter */
      command = gtk_editable_get_chars (GTK_EDITABLE (guilebuffer), 0,
					gtk_text_get_length (GTK_TEXT (guilebuffer)));
      gh_eval_str_with_catch (command, err_handler);
      fflush (stdout);
      g_free (command);
      break;

    case 'c':
      gtk_widget_destroy (widget);
      guilebuffer = NULL;
    }
  }

  return FALSE;
}


static void init_guile_hiliting (void)
{
  GdkColor red, brownish, blueish;
  static GList *stentries = NULL;
  static GList *pentries = NULL;
  GdkColormap *cmap;
  GdkFont *commentfont;
  GdkFont *keywordsfont;

  if (!stentries) {
  
    cmap = gtk_widget_get_colormap (GTK_WIDGET (mainbuffer));
    commentfont = gdk_font_load ("-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*");
    keywordsfont = gdk_font_load ("-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*");

    red.pixel = 0;
    red.red = 65535;
    red.green = 0;
    red.blue = 0;
    
    brownish.pixel = 0;
    brownish.red = 65535;
    brownish.green = 43908;
    brownish.blue = 16383;

    blueish.pixel = 0;
    blueish.red = 40000;
    blueish.green = 0;
    blueish.blue = 55000;
    
    gdk_color_alloc(cmap, &red);
    gdk_color_alloc(cmap, &brownish);
    gdk_color_alloc (cmap, &blueish);

    stentries = gtk_editor_stentry_new ("Scheme-comment",
					";", "\n", FALSE,
					commentfont, &red, NULL,
					gtk_editor_stentry_new ("string",
								"\"", "\"", FALSE,
								NULL, &brownish, NULL,
								NULL));
    pentries = gtk_editor_pentry_new ("scheme-keywords",
		      "\\b\\(define\\|if\\|or\\|let\\*?\\)\\b",
				      keywordsfont, &blueish, NULL, NULL);
  }

  gtk_editor_install_stable (guilebuffer, stentries);
  gtk_editor_install_patterns (guilebuffer, pentries);
}

static void
pop_guile_buffer (void)
{
  GtkWidget *window;
  GtkWidget *editor;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Guile Buffer Test");

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (gtk_widget_destroy),
		      NULL);
  editor = gtk_editor_new (NULL, NULL);

  gtk_widget_set_usize (editor, 400, 150);
  gtk_container_add (GTK_CONTAINER (window), editor);
  gtk_widget_show (editor);

  gtk_text_set_editable (GTK_TEXT (editor), TRUE);

  gtk_signal_connect_object (GTK_OBJECT (editor), "key_press_event",
		      GTK_SIGNAL_FUNC (guile_buffer_key_press),
		      GTK_OBJECT (window));

  guilebuffer = GTK_EDITOR (editor);

  init_guile_hiliting ();
  gtk_widget_show (window);
  gtk_widget_grab_focus (GTK_WIDGET (guilebuffer));
}

#endif

/* key events */
static gint
mainbuf_key_press (GtkWidget *widget, GdkEventKey *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  switch (event->keyval) {
  case GDK_F1:
    clear_buffer ();
    break;
  case GDK_F2:
    load_file ();
    break;
  case GDK_F3:
    save_file ();
    break;
  case GDK_F4:
    search_from_point ();
    break;
  case GDK_F5:
    search_back_from_point ();
    break;
  case GDK_F6:
    regex_search_from_point ();
    break;
  case GDK_F7:
    regex_search_back_from_point ();
    break;

  case 'l':
    if (event->state & GDK_CONTROL_MASK) {
      gtk_editor_hilite_screen (mainbuffer);
    }

    /* ctrl stuff */
#ifdef HAVE_LIBGUILE
  case 'g':
    if (event->state & GDK_CONTROL_MASK) {
      if (!guilebuffer) {
	pop_guile_buffer ();
      }
    }
#endif

  }

  return FALSE;
}

/* syntax table test */
static void
install_st (void)
{
  GdkColor red, brownish;
  GList *entries;
  GdkColormap *cmap = gtk_widget_get_colormap (GTK_WIDGET (mainbuffer));
  GdkFont *commentfont =
    gdk_font_load ("-adobe-courier-medium-i-normal-*-12-*-*-*-*-*-*-*");

  red.pixel = 0;
  red.red = 65535;
  red.green = 0;
  red.blue = 0;

  brownish.pixel = 0;
  brownish.red = 65535;
  brownish.green = 43908;
  brownish.blue = 16383;

  gdk_color_alloc(cmap, &red);
  gdk_color_alloc(cmap, &brownish);

  entries = gtk_editor_stentry_new ("C-comment",
				    "/\\*", "\\*/", FALSE,
				    commentfont, &red, NULL,
				    gtk_editor_stentry_new ("C++-comment",
							    "//", "\n", FALSE,
							    commentfont, &red, NULL,
				    gtk_editor_stentry_new ("string",
							    "\"", "\"", FALSE,
							    NULL, &brownish, NULL,
				    gtk_editor_stentry_new ("chars",
							    "'", "'", FALSE,
							    NULL, &brownish, NULL,
							    NULL))));

  gtk_editor_install_stable (mainbuffer, entries);
  gtk_editor_free_stentries (entries);
  gdk_font_unref (commentfont);
}

/* hilite pattern test */
static void
install_pat (void)
{
  GdkColor blue, greenish, blueish;
  GList *entries;
  GdkColormap *cmap = gtk_widget_get_colormap (GTK_WIDGET (mainbuffer));
  GdkFont *keywordsfont =
    gdk_font_load ("-adobe-courier-bold-r-normal-*-12-*-*-*-*-*-*-*");

  blue.pixel = 0;
  blue.red = 0;
  blue.green = 0;
  blue.blue = 65535;

  greenish.pixel = 0;
  greenish.red = 30000;
  greenish.green = 45000;
  greenish.blue = 0;

  blueish.pixel = 0;
  blueish.red = 40000;
  blueish.green = 0;
  blueish.blue = 55000;

  gdk_color_alloc (cmap, &blue);
  gdk_color_alloc (cmap, &greenish);
  gdk_color_alloc (cmap, &blueish);

  entries = gtk_editor_pentry_new ("keywords",
				   "\\b\\(while\\|for\\|if\\|else\\)\\b",
				   keywordsfont, &blue, NULL,
            gtk_editor_pentry_new ("types",
				   "\\b\\(int\\|float\\|char\\|void\\)\\b",
				   NULL, &greenish, NULL, 
            gtk_editor_pentry_new ("cpp",
		   "^#[ \t]*\\(if\\|ifdef\\|else\\|define\\|endif\\)\\b",
				   NULL, &blueish, NULL, NULL)));

  gtk_editor_install_patterns (mainbuffer, entries);
  gtk_editor_free_pentries (entries);
  gdk_font_unref (keywordsfont);
}


static void
quit (void)
{
  gtk_exit (0);
}

/* main program */

static GtkWidget*
create_window ()
{
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *editor;
  GtkWidget *entry;
  GtkHint *hint;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Editor Test");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  editor = gtk_editor_new(NULL, NULL);
  gtk_widget_set_usize (editor, 450, 250);
  gtk_widget_show (editor);

  hint = gtk_hint_new ();
  gtk_hint_set_hint (hint, editor,
		     "\\<\\(\\w[a-zA-Z0-9_]*\\)[\\t\\n ]*([^()]*$",
		     hint_callback, NULL);

  gtk_text_set_editable (GTK_TEXT (editor), TRUE);
  gtk_box_pack_start (GTK_BOX (vbox), editor, TRUE, TRUE, 0);

  entry = gtk_entry_new ();
  gtk_widget_show (entry);
  gtk_box_pack_start (GTK_BOX (vbox), entry, FALSE, FALSE, 0);

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC (quit),
		      NULL);

  minibuffer = GTK_ENTRY (entry);
  gtk_signal_connect (GTK_OBJECT (minibuffer), "key_press_event",
		      GTK_SIGNAL_FUNC (minibuf_key_press),
		      NULL);
  mainbuffer = GTK_EDITOR (editor);
  gtk_signal_connect (GTK_OBJECT (mainbuffer), "key_press_event",
		      GTK_SIGNAL_FUNC (mainbuf_key_press),
		      NULL);

  return window;
}

#ifdef HAVE_LIBGUILE
void
main_prog (int argc, char *argv[])
{
  gtk_editor_guile_init ();
  gtk_editor_guile_set_default (mainbuffer);
  gtk_editor_guile_install_hilite_functions ();
  gtk_editor_guile_install_search_functions ();
  gtk_editor_guile_install_editing_functions ();
  gtk_editor_guile_install_movement_functions ();

  gtk_main ();
}
#endif

int
main (int argc, char *argv[]) 
{
  GtkWidget *main_window;

  /* initialize */
  gtk_init (&argc, &argv);
  gtk_rc_parse ("editorrc");

  main_window = create_window ();

  /* show */
  gtk_widget_show (main_window);

  gtk_widget_grab_focus (GTK_WIDGET (mainbuffer));

  install_st ();
  install_pat ();

  hint_load_db ("GTAGS");

  /* args? */
  if (argc > 1) {
    load_file_cb (g_strdup (argv[1])); /* we need a copy 'cause load
					* frees the string */
  }
  
  /* main loop */
#ifdef HAVE_LIBGUILE
  gh_enter (argc, argv, main_prog);
#else
  gtk_main ();
#endif /* HAVE_LIBGUILE */

  return EXIT_SUCCESS;
}
